@@ -45,8 +45,9 @@ int tcmu_emulate_set_tgt_port_grps(struct tcmu_device *dev,
 struct tgt_port *tcmu_get_enabled_port(struct list_head *);
 int tcmu_get_alua_grps(struct tcmu_device *, struct list_head *);
 void tcmu_release_alua_grps(struct list_head *);
-int alua_implicit_transition(struct tcmu_device *dev, struct tcmulib_cmd *cmd);
+int alua_implicit_transition(struct tcmu_device *dev, struct tcmulib_cmd *cmd,
+			     bool is_read);
 bool lock_is_required(struct tcmu_device *dev);
-int alua_check_state(struct tcmu_device *dev, struct tcmulib_cmd *cmd);
+int alua_check_state(struct tcmu_device *dev, struct tcmulib_cmd *cmd, bool is_read);
 
 #endif
