@@ -16,14 +16,19 @@
 
 struct tcmu_config {
 	pthread_t thread_id;
-	char *path;
 
 	int log_level;
-	char *log_dir_path;
+	int def_log_level;
+
+	char log_dir[PATH_MAX];
+	char def_log_dir[PATH_MAX];
+
+	struct tcmulib_context *ctx;
 };
 
 /*
- * There are 5 logging levels supported in tcmu.conf:
+ * There are 6 logging levels supported in tcmu.conf:
+ *    0: CRIT
  *    1: ERROR
  *    2: WARNING
  *    3: INFO
@@ -31,7 +36,8 @@ struct tcmu_config {
  *    5: DEBUG SCSI CMD
  */
 enum {
-	TCMU_CONF_LOG_LEVEL_MIN = 1,
+	TCMU_CONF_LOG_LEVEL_MIN = 0,
+	TCMU_CONF_LOG_CRIT = 0,
 	TCMU_CONF_LOG_ERROR = 1,
 	TCMU_CONF_LOG_WARN,
 	TCMU_CONF_LOG_INFO,
@@ -41,6 +47,7 @@ enum {
 };
 
 static const char *const log_level_lookup[] = {
+	[TCMU_CONF_LOG_CRIT]	= "CRIT",
 	[TCMU_CONF_LOG_ERROR]	= "ERROR",
 	[TCMU_CONF_LOG_WARN]	= "WARNING",
 	[TCMU_CONF_LOG_INFO]	= "INFO",
@@ -48,8 +55,9 @@ static const char *const log_level_lookup[] = {
 	[TCMU_CONF_LOG_DEBUG_SCSI_CMD]	= "DEBUG SCSI CMD",
 };
 
+struct tcmu_config* tcmu_initialize_config(void);
 void tcmu_free_config(struct tcmu_config *cfg);
-struct tcmu_config * tcmu_parse_config(const char *path);
+int tcmu_load_config(struct tcmu_config *cfg);
 int tcmu_watch_config(struct tcmu_config *cfg);
 void tcmu_unwatch_config(struct tcmu_config *cfg);
 
