@@ -19,11 +19,7 @@
 #include <gio/gio.h>
 #include <pthread.h>
 
-#include "scsi_defs.h"
 #include "darray.h"
-#include "ccan/list/list.h"
-#include "tcmur_aio.h"
-#include "tcmu-runner.h"
 
 #define KERN_IFACE_VER 2
 
@@ -49,6 +45,7 @@ struct tcmu_device {
 
 	uint64_t num_lbas;
 	uint32_t block_size;
+	uint32_t block_size_shift;
 	uint32_t max_xfer_len;
 	uint32_t opt_xcopy_rw_len;
 	bool split_unmaps;
@@ -57,6 +54,7 @@ struct tcmu_device {
 	uint32_t unmap_gran_align;
 	unsigned int write_cache_enabled:1;
 	unsigned int solid_state_media:1;
+	unsigned int unmap_enabled:1;
 
 	char dev_name[16]; /* e.g. "uio14" */
 	char tcm_hba_name[16]; /* e.g. "user_8" */
@@ -66,7 +64,6 @@ struct tcmu_device {
 	struct tcmulib_handler *handler;
 	struct tcmulib_context *ctx;
 
-	void *d_private; /* private ptr for the daemon */
 	void *hm_private; /* private ptr for handler module */
 };
 
