@@ -365,7 +365,7 @@ static int align_and_split_unmap(struct tcmu_device *dev,
 		pthread_mutex_lock(&state->lock);
 		state->refcount++;
 		pthread_mutex_unlock(&state->lock);
-	
+
 		ret = aio_request_schedule(dev, tcmur_ucmd, unmap_work_fn,
 					   tcmur_cmd_complete);
 		if (ret != TCMU_STS_ASYNC_HANDLED)
@@ -373,9 +373,7 @@ static int align_and_split_unmap(struct tcmu_device *dev,
 
 		nlbas -= lbas;
 		lba += lbas;
-
 		lbas = min(opt_unmap_gran, nlbas);
-
 	}
 
 	return ret;
@@ -682,9 +680,19 @@ static int handle_unmap_in_writesame(struct tcmu_device *dev,
 	uint8_t *cdb = cmd->cdb;
 	uint64_t lba = tcmu_cdb_get_lba(cdb);
 	uint64_t nlbas = tcmu_cdb_get_xfer_length(cdb);
+	uint32_t align = tcmu_dev_get_unmap_gran_align(dev);
 	struct unmap_state *state;
 	int ret;
 
+	/* If not aligned then falls back to the writesame without unmap */
+	if (lba % align || nlbas % align) {
+		tcmu_dev_dbg(dev,
+			     "Start lba: %"PRIu64" or nlbas: %"PRIu64" not aligned to %"PRIu32"\n",
+			     lba, nlbas, align);
+		tcmu_dev_dbg(dev, "Falls back to writesame without unmap!\n");
+		return TCMU_STS_NOT_HANDLED;
+	}
+
 	tcmu_dev_dbg(dev, "Do UNMAP in WRITE_SAME cmd!\n");
 
 	ret = unmap_init(dev, cmd);
@@ -703,6 +711,23 @@ static int handle_unmap_in_writesame(struct tcmu_device *dev,
 	return TCMU_STS_ASYNC_HANDLED;
 }
 
+static int tcmur_writesame_work_fn(struct tcmu_device *dev, void *data)
+{
+	struct tcmur_cmd *tcmur_cmd = data;
+	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
+	tcmur_writesame_fn_t write_same_fn = tcmur_cmd->cmd_state;
+	uint8_t *cdb = cmd->cdb;
+	uint64_t off = tcmu_cdb_to_byte(dev, cdb);
+	uint64_t len = tcmu_lba_to_byte(dev, tcmu_cdb_get_xfer_length(cdb));
+
+	/*
+	 * Write contents of the logical block data(from the Data-Out Buffer)
+	 * to each LBA in the specified LBA range.
+	 */
+	return write_same_fn(dev, tcmur_cmd, off, len, cmd->iovec,
+			     cmd->iov_cnt);
+}
+
 static int handle_writesame(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 {
 	struct tcmur_handler *rhandler = tcmu_get_runner_handler(dev);
@@ -716,12 +741,30 @@ static int handle_writesame(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	struct write_same *write_same;
 	int i, ret;
 
+	if (tcmu_dev_in_recovery(dev))
+		return TCMU_STS_BUSY;
+
+	ret = alua_check_state(dev, cmd, false);
+	if (ret)
+		return ret;
+
 	ret = handle_writesame_check(dev, cmd);
 	if (ret)
 		return ret;
 
-	if (rhandler->unmap && (cmd->cdb[1] & 0x08))
-		return handle_unmap_in_writesame(dev, cmd);
+	if (rhandler->unmap && (cmd->cdb[1] & 0x08)) {
+		ret = handle_unmap_in_writesame(dev, cmd);
+		if (ret != TCMU_STS_NOT_HANDLED)
+			return ret;
+	}
+
+	if (rhandler->writesame) {
+		tcmur_cmd->cmd_state = rhandler->writesame;
+		tcmur_cmd->done = handle_generic_cbk;
+		return aio_request_schedule(dev, tcmur_cmd,
+					    tcmur_writesame_work_fn,
+					    tcmur_cmd_complete);
+	}
 
 	max_xfer_length = tcmu_dev_get_max_xfer_len(dev) * block_size;
 	length = round_up(length, max_xfer_length);
@@ -749,51 +792,6 @@ static int handle_writesame(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 				    tcmur_cmd_complete);
 }
 
-static int tcmur_writesame_work_fn(struct tcmu_device *dev, void *data)
-{
-	struct tcmur_cmd *tcmur_cmd = data;
-	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
-	tcmur_writesame_fn_t write_same_fn = tcmur_cmd->cmd_state;
-	uint8_t *cdb = cmd->cdb;
-	uint64_t off = tcmu_cdb_to_byte(dev, cdb);
-	uint64_t len = tcmu_lba_to_byte(dev, tcmu_cdb_get_xfer_length(cdb));
-
-	/*
-	 * Write contents of the logical block data(from the Data-Out Buffer)
-	 * to each LBA in the specified LBA range.
-	 */
-	return write_same_fn(dev, tcmur_cmd, off, len, cmd->iovec,
-			     cmd->iov_cnt);
-}
-
-int tcmur_handle_writesame(struct tcmu_device *dev, struct tcmur_cmd *tcmur_cmd,
-			   tcmur_writesame_fn_t write_same_fn)
-{
-	struct tcmur_handler *rhandler = tcmu_get_runner_handler(dev);
-	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
-	int ret;
-
-	if (tcmu_dev_in_recovery(dev))
-		return TCMU_STS_BUSY;
-
-	ret = alua_check_state(dev, cmd);
-	if (ret)
-		return ret;
-
-	ret = handle_writesame_check(dev, cmd);
-	if (ret)
-		return ret;
-
-	if (rhandler->unmap && (cmd->cdb[1] & 0x08))
-		return handle_unmap_in_writesame(dev, cmd);
-
-	tcmur_cmd->cmd_state = write_same_fn;
-	tcmur_cmd->done = handle_generic_cbk;
-
-	return aio_request_schedule(dev, tcmur_cmd, tcmur_writesame_work_fn,
-				    tcmur_cmd_complete);
-}
-
 /* async write verify */
 struct write_verify_state {
 	struct iovec *w_iovec;
@@ -1161,6 +1159,12 @@ static int xcopy_parse_target_descs(struct tcmu_device *udev,
 {
 	int i, ret;
 
+	if (tdll % XCOPY_TARGET_DESC_LEN) {
+		tcmu_dev_err(udev,
+			"CSCD descriptor list length %u not a multiple of %u\n",
+			(unsigned int)tdll, XCOPY_TARGET_DESC_LEN);
+		return TCMU_STS_NOTSUPP_TGT_DESC_TYPE;
+	}
 	/* From spc4r36q,section 6.4.3.4 CSCD DESCRIPTOR LIST LENGTH field
 	 * If the number of CSCD descriptors exceeds the allowed number, the copy
 	 * manager shall terminate the command with CHECK CONDITION status, with
@@ -1173,7 +1177,7 @@ static int xcopy_parse_target_descs(struct tcmu_device *udev,
 		return TCMU_STS_TOO_MANY_TGT_DESC;
 	}
 
-	for (i = 0; i < RCR_OP_MAX_TARGET_DESC_COUNT; i++) {
+	for (i = 0; tdll >= XCOPY_TARGET_DESC_LEN; i++) {
 		/*
 		 * Only Identification Descriptor Target Descriptor support
 		 * for now.
@@ -1184,6 +1188,7 @@ static int xcopy_parse_target_descs(struct tcmu_device *udev,
 				return ret;
 
 			tgt_desc += XCOPY_TARGET_DESC_LEN;
+			tdll -= XCOPY_TARGET_DESC_LEN;
 		} else {
 			tcmu_dev_err(udev, "Unsupport target descriptor type code 0x%x\n",
 				     tgt_desc[0]);
@@ -1191,6 +1196,7 @@ static int xcopy_parse_target_descs(struct tcmu_device *udev,
 		}
 	}
 
+	ret = TCMU_STS_CP_TGT_DEV_NOTCONN;
 	if (xcopy->src_dev)
 		ret = xcopy_locate_udev(udev->ctx, xcopy->dst_tid_wwn,
 					&xcopy->dst_dev);
@@ -1308,6 +1314,12 @@ static int xcopy_parse_parameter_list(struct tcmu_device *dev,
 	 * data, after the last segment descriptor.
 	 * */
 	inline_dl = be32toh(*(uint32_t *)&par[12]);
+	if (inline_dl != 0) {
+		tcmu_dev_err(dev, "non-zero xcopy inline_dl %u unsupported\n",
+			     inline_dl);
+		ret = TCMU_STS_INVALID_PARAM_LIST_LEN;
+		goto err;
+	}
 
 	/* From spc4r31, section 6.3.1 EXTENDED COPY command introduction
 	 *
@@ -1349,6 +1361,18 @@ static int xcopy_parse_parameter_list(struct tcmu_device *dev,
 	if (ret != TCMU_STS_OK)
 		goto err;
 
+	/*
+	 * tcmu-runner can't determine whether the device(s) referred to in an
+	 * XCOPY request should be accessible to the initiator via transport
+	 * settings, ACLs, etc. XXX Consequently, we need to fail any
+	 * cross-device requests for safety reasons.
+	 */
+	if (dev != xcopy->src_dev || dev != xcopy->dst_dev) {
+		tcmu_dev_err(dev, "Cross-device XCOPY not supported\n");
+		ret = TCMU_STS_CP_TGT_DEV_NOTCONN;
+		goto err;
+	}
+
 	if (tcmu_dev_get_block_size(xcopy->src_dev) !=
 	    tcmu_dev_get_block_size(xcopy->dst_dev)) {
 		tcmu_dev_err(dev, "The block size of src dev %u != dst dev %u\n",
@@ -1649,14 +1673,33 @@ static int handle_caw_check(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	return TCMU_STS_OK;
 }
 
+static int tcmur_caw_fn(struct tcmu_device *dev, void *data)
+{
+	struct tcmur_cmd *tcmur_cmd = data;
+	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
+	tcmur_caw_fn_t caw_fn = tcmur_cmd->cmd_state;
+	uint64_t off = tcmu_cdb_to_byte(dev, cmd->cdb);
+	size_t half = (tcmu_iovec_length(cmd->iovec, cmd->iov_cnt)) / 2;
+
+	return caw_fn(dev, tcmur_cmd, off, half, cmd->iovec, cmd->iov_cnt);
+}
+
 static int handle_caw(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 {
+	struct tcmur_handler *rhandler = tcmu_get_runner_handler(dev);
 	struct tcmur_cmd *tcmur_cmd = cmd->hm_private;
 	size_t half = (tcmu_iovec_length(cmd->iovec, cmd->iov_cnt)) / 2;
 	struct tcmur_device *rdev = tcmu_dev_get_private(dev);
 	uint8_t sectors = cmd->cdb[13];
 	int ret;
 
+	if (tcmu_dev_in_recovery(dev))
+		return TCMU_STS_BUSY;
+
+	ret = alua_check_state(dev, cmd, false);
+	if (ret)
+		return ret;
+
 	/* From sbc4r12a section 5.3 COMPARE AND WRITE command
 	 * A NUMBER OF LOGICAL BLOCKS field set to zero specifies that no
 	 * read operations shall be performed, no logical block data shall
@@ -1673,6 +1716,13 @@ static int handle_caw(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	if (ret)
 		return ret;
 
+	if (rhandler->caw) {
+		tcmur_cmd->cmd_state = rhandler->caw;
+		tcmur_cmd->done = handle_generic_cbk;
+		return aio_request_schedule(dev, tcmur_cmd, tcmur_caw_fn,
+					    tcmur_cmd_complete);
+	}
+
 	if (tcmur_cmd_state_init(tcmur_cmd, 0, half))
 		return TCMU_STS_NO_RESOURCE;
 
@@ -1690,54 +1740,6 @@ static int handle_caw(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	return ret;
 }
 
-static int tcmur_caw_fn(struct tcmu_device *dev, void *data)
-{
-	struct tcmur_cmd *tcmur_cmd = data;
-	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
-	tcmur_caw_fn_t caw_fn = tcmur_cmd->cmd_state;
-	uint64_t off = tcmu_cdb_to_byte(dev, cmd->cdb);
-	size_t half = (tcmu_iovec_length(cmd->iovec, cmd->iov_cnt)) / 2;
-
-	return caw_fn(dev, tcmur_cmd, off, half, cmd->iovec, cmd->iov_cnt);
-}
-
-int tcmur_handle_caw(struct tcmu_device *dev, struct tcmur_cmd *tcmur_cmd,
-		     tcmur_caw_fn_t caw_fn)
-{
-	struct tcmulib_cmd *cmd = tcmur_cmd->lib_cmd;
-	uint8_t sectors = cmd->cdb[13];
-	int ret;
-
-	if (tcmu_dev_in_recovery(dev))
-		return TCMU_STS_BUSY;
-
-	/* From sbc4r12a section 5.3 COMPARE AND WRITE command
-	 * A NUMBER OF LOGICAL BLOCKS field set to zero specifies that no
-	 * read operations shall be performed, no logical block data shall
-	 * be transferred from the Data-Out Buffer, no compare operations
-	 * shall be performed, and no write operations shall be performed.
-	 * This condition shall not be considered an error.
-	 */
-	if (!sectors) {
-		tcmu_dev_dbg(dev, "NUMBER OF LOGICAL BLOCKS is zero, just return ok.\n");
-		return TCMU_STS_OK;
-	}
-
-	ret = alua_check_state(dev, cmd);
-	if (ret)
-		return ret;
-
-	ret = handle_caw_check(dev, cmd);
-	if (ret)
-		return ret;
-
-	tcmur_cmd->cmd_state = caw_fn;
-	tcmur_cmd->done = handle_generic_cbk;
-
-	return aio_request_schedule(dev, tcmur_cmd, tcmur_caw_fn,
-				    tcmur_cmd_complete);
-}
-
 /* async flush */
 static int flush_work_fn(struct tcmu_device *dev, void *data)
 {
@@ -2118,6 +2120,7 @@ static int tcmur_cmd_handler(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	struct tcmur_handler *rhandler = tcmu_get_runner_handler(dev);
 	struct tcmur_device *rdev = tcmu_dev_get_private(dev);
 	uint8_t *cdb = cmd->cdb;
+	bool is_read = false;
 
 	track_aio_request_start(rdev);
 
@@ -2128,10 +2131,12 @@ static int tcmur_cmd_handler(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 
 	/* Don't perform alua implicit transition if command is not supported */
 	switch(cdb[0]) {
+	/* Skip to grab the lock for reads */
 	case READ_6:
 	case READ_10:
 	case READ_12:
 	case READ_16:
+		is_read = true;
 	case WRITE_6:
 	case WRITE_10:
 	case WRITE_12:
@@ -2146,7 +2151,7 @@ static int tcmur_cmd_handler(struct tcmu_device *dev, struct tcmulib_cmd *cmd)
 	case WRITE_SAME:
 	case WRITE_SAME_16:
 	case FORMAT_UNIT:
-		ret = alua_check_state(dev, cmd);
+		ret = alua_check_state(dev, cmd, is_read);
 		if (ret)
 			goto untrack;
 		break;
