@@ -24,17 +24,14 @@ int tcmur_cmd_passthrough_handler(struct tcmu_device *dev, struct tcmulib_cmd *c
 bool tcmur_handler_is_passthrough_only(struct tcmur_handler *rhandler);
 void tcmur_tcmulib_cmd_complete(struct tcmu_device *dev,
 				struct tcmulib_cmd *cmd, int ret);
+
 typedef int (*tcmur_writesame_fn_t)(struct tcmu_device *dev,
 				    struct tcmur_cmd *tcmur_cmd, uint64_t off,
 				    uint64_t len, struct iovec *iov,
 				    size_t iov_cnt);
-int tcmur_handle_writesame(struct tcmu_device *dev, struct tcmur_cmd *tcmur_cmd,
-			   tcmur_writesame_fn_t write_same_fn);
 
 typedef int (*tcmur_caw_fn_t)(struct tcmu_device *dev,
 			      struct tcmur_cmd *tcmur_cmd, uint64_t off,
 			      uint64_t len, struct iovec *iov, size_t iov_cnt);
-int tcmur_handle_caw(struct tcmu_device *dev, struct tcmur_cmd *tcmur_cmd,
-                     tcmur_caw_fn_t caw_fn);
 
 #endif /* __TCMUR_CMD_HANDLER_H */
